"""
Py-ART: The Python ARM Radar Toolkit
=====================================

"""

# print information on citing Py-ART, this message can be suppressed by
# setting the PYART_QUIET environment variable
_citation_text = """
## You are using the Python ARM Radar Toolkit (Py-ART), an open source
## library for working with weather radar data. Py-ART is partly
## supported by the U.S. Department of Energy as part of the Atmospheric
## Radiation Measurement (ARM) Climate Research Facility, an Office of
## Science user facility.
##
## If you use this software to prepare a publication, please cite:
##
##     JJ Helmus and SM Collis, JORS 2016, doi: 10.5334/jors.119
"""
from os import environ as _environ
if 'PYART_QUIET' not in _environ:
    print(_citation_text)

# Make sure that deprecation warnings get printed by default
import warnings as _warnings
_warnings.simplefilter("always", DeprecationWarning)

# print out helpful message if build fails or importing from source tree
from . import __check_build

# import subpackages
from . import core
from . import io
from . import correct
from . import graph
from . import map
from . import filters
from . import util
from . import testing
from . import config
from . import aux_io
from . import retrieve
from . import bridge

# root level functions
from .config import load_config
from ._debug_info import _debug_info

import os.path as _osp
import functools as _functools
from pkg_resources import DistributionNotFound, get_distribution
import sys as _sys

# Get the version
try:
    __version__ = get_distribution("arm_pyart").version
except DistributionNotFound:
    # package is not installed
    __version__ = '0.0.0'



"""
=================================
Create a PPI plot from a MDV file
=================================

An example which creates a PPI plot of a MDV file using a RadarDisplay object.

"""
print(__doc__)

# Author: Jonathan J. Helmus (jhelmus@anl.gov)
# License: BSD 3 clause

import matplotlib.pyplot as plt
import pyart
from pyart.testing import get_test_data

filename = get_test_data('110635.mdv')

# create the plot using RadarDisplay
radar = pyart.io.read_mdv(filename)
display = pyart.graph.RadarDisplay(radar)
fig = plt.figure(figsize=[5, 5])
ax = fig.add_subplot(111)
display.plot('reflectivity', 0, vmin=-16., vmax=64, title="PPI",
              cmap='pyart_HomeyerRainbow')
display.set_limits(ylim=[-150, 150], xlim=[-150, 150])
plt.show()



"""
======================================
Create a PPI plot from a Cfradial file
======================================

An example which creates a PPI plot of a Cfradial file.

"""
print(__doc__)

# Author: Max Grover (mgrover@anl.gov)
# License: BSD 3 clause

import matplotlib.pyplot as plt
import pyart
from pyart.testing import get_test_data

# Locate the test data and read in using main read method
filename = get_test_data('swx_20120520_0641.nc')
radar = pyart.io.read(filename)

# Setup the display, which automatically detects this is a ppi scan
display = pyart.graph.RadarDisplay(radar)
fig = plt.figure()
ax = fig.add_subplot(111)
display.plot('reflectivity_horizontal', 0, vmin=-32, vmax=64.)
display.plot_range_rings([10, 20, 30, 40])
display.plot_cross_hair(5.)
plt.show()



"""
A script for comparing the e Vulpiani, Maesaka and Schneedbeli methods
for KDP estimation with sample data.
"""

import numpy as np
import matplotlib.pyplot as plt
import pytest

from pyart.retrieve.kdp_proc import kdp_maesaka, kdp_vulpiani, kdp_schneebeli
from pyart.testing import sample_objects
from pyart.config import get_field_name


def _make_real_psidp_radar():
    """
    Create single-ray radar with linear differential phase profile with
    specified slope.
    ---
    Returns
    -------
    radar : Radar
            PyART radar instance with differential phase profile in deg.
    """
    psidp = np.array([[-2.33313751e+00, 1.80617523e+00, 7.17742920e-01,
                     1.82811661e+01, 1.89352417e+01, 1.67904205e+01]])
    psidp = np.ma.array(psidp)
    radar = sample_objects.make_empty_ppi_radar(len(psidp[0]), 1, 1)
    psidp_dict = {
                  'data': psidp,
                  }
    radar.add_field(get_field_name('differential_phase'), psidp_dict)
    # Define real ranges
    radar.range['data'] = 75*np.arange(0, len(psidp[0]))
    return radar


@pytest.mark.mpl_image_compare(tolerance=30)
def test_compare_kdp_estimation_methods():
    # Get profile of noisy psidp
    prof_psidp = _make_real_psidp_radar()
    # Maesaka method
    kdp_mae, phidpf_mae, phidp_mae = kdp_maesaka(prof_psidp,
                                                 maxiter=1000,
                                                 check_outliers=False,
                                                 parallel=False)

    # Vulpiani method (note windsize is just a guess here..)
    kdp_vulp, phidp_vulp = kdp_vulpiani(prof_psidp,
                                        windsize=2,
                                        n_iter=20,
                                        parallel=False,
                                        band='X')
    # Kalman filter method
    kdp_schnee, kdp_std_schnee, phidp_schnee = kdp_schneebeli(prof_psidp,
                                                              parallel=False,
                                                              band='X')
    # Create figure
    fig = plt.figure(figsize=(10, 10))
    plt.subplot(2, 1, 1)
    plt.grid(True)
    plt.title('Kdp estimation')
    ranges = prof_psidp.range['data']
    plt.plot(ranges, kdp_mae['data'][0])
    plt.plot(ranges, kdp_vulp['data'][0])
    plt.plot(ranges, kdp_schnee['data'][0])
    plt.xlabel('Range [m]')
    plt.ylabel('Kdp [deg/km]')
    plt.legend(['Maesaka', 'Vulpiani', 'Schneebeli'], loc=0)
    plt.subplot(2, 1, 2)
    plt.grid(True)
    plt.title('Reconstructed Phidp')
    ranges = prof_psidp.range['data']
    phidp_mae = 0.5 * (phidp_mae['data'][0] + phidp_mae['data'][0])
    plt.plot(ranges, phidp_mae)
    plt.plot(ranges, phidp_vulp['data'][0])
    plt.plot(ranges, phidp_schnee['data'][0])
    plt.plot(ranges, prof_psidp.fields['differential_phase']['data'][0])
    plt.xlabel('Range [m]')
    plt.ylabel('Diff. phase [deg]')
    plt.legend(['Maesaka', 'Vulpiani', 'Schneebeli', 'Real Psidp'], loc=0)
    try:
        return fig
    finally:
        plt.close(fig)



"""
=================================
Plot a Cross Section from a Grid
=================================

This is an example of how to plot a cross section
of your radar grid using the GridMapDisplay

"""
print(__doc__)

# Author: Max Grover (mgrover@anl.gov)
# License: BSD 3 clause

import matplotlib.pyplot as plt
import cartopy.crs as ccrs

import pyart
from pyart.testing import get_test_data

# Read in the data from two XSAPR radars
xsapr_sw_file = get_test_data('swx_20120520_0641.nc')
xsapr_se_file = get_test_data('sex_20120520_0641.nc')
radar_sw = pyart.io.read_cfradial(xsapr_sw_file)
radar_se = pyart.io.read_cfradial(xsapr_se_file)

# Filter out gates with reflectivity > 100 from both radars
gatefilter_se = pyart.filters.GateFilter(radar_se)
gatefilter_se.exclude_transition()
gatefilter_se.exclude_above('corrected_reflectivity_horizontal', 100)
gatefilter_sw = pyart.filters.GateFilter(radar_sw)
gatefilter_sw.exclude_transition()
gatefilter_sw.exclude_above('corrected_reflectivity_horizontal', 100)

# perform Cartesian mapping, limit to the reflectivity field.
grid = pyart.map.grid_from_radars(
    (radar_se, radar_sw), gatefilters=(gatefilter_se, gatefilter_sw),
    grid_shape=(20, 181, 181),
    grid_limits=((500, 10000), (-50000, 40000), (-60000, 40000)),
    grid_origin = (36.57861, -97.363611),
    fields=['corrected_reflectivity_horizontal'])

# Define some start and end points, using (latitude, longitude)
start = (36.7, -97.7)
end = (36.2, -97.8)

# Setup the figure, and plot our x/y view of the radar
fig = plt.figure(figsize=(18,6))
ax1 = plt.subplot(121, projection=ccrs.PlateCarree())
display = pyart.graph.GridMapDisplay(grid)
display.plot_grid("corrected_reflectivity_horizontal", ax=ax1,
                  cmap='pyart_HomeyerRainbow', vmin=-20, vmax=70)

# Plot our start and end points, as well as a line in between the two
ax1.scatter(start[1], start[0], color='tab:blue', label='Start')
ax1.scatter(end[1], end[0], color='black', label='End')
ax1.plot([start[1], end[1]], [start[0], end[0]], color='k', linestyle=':')
plt.legend(loc='upper right')

# Add a cross section, using our start and end points, and set our x-axis as latitude (lat)
ax2 = plt.subplot(122)
display.plot_cross_section("corrected_reflectivity_horizontal", start, end, x_axis='lat',
                           cmap='pyart_HomeyerRainbow', vmin=-20, vmax=70)



"""
======================================
Reading NEXRAD Data from the AWS Cloud
======================================

Within this example, we show how you can remotely access Next Generation Weather Radar (NEXRAD) Data from Amazon Web Services
and plot quick looks of the datasets.

"""
print(__doc__)

# Author: Max Grover (mgrover@anl.gov)
# License: BSD 3 clause

import cartopy.crs as ccrs
import matplotlib.pyplot as plt

import pyart

######################################%%
# Read NEXRAD Level 2 Data
# ------------------------
#
# Let's start first with NEXRAD Level 2 data, which is ground-based radar data collected
# by the National Oceanic and Atmospheric Administration (NOAA), as a part of the National Weather Service
# ### Configure our Filepath for NEXRAD Level 2 Data
# We will access data from the **noaa-nexrad-level2** bucket, with the data organized as:
# 
#    ``s3://noaa-nexrad-level2/year/month/date/radarsite/{radarsite}{year}{month}{date}_{hour}{minute}{second}_V06``
# 
# Where in our case, we are using a sample data file from Houston, Texas (KHGX)
# on March 22, 2022, at 1201:25 UTC. This means our path would look like:

aws_nexrad_level2_file = "s3://noaa-nexrad-level2/2022/03/22/KHGX/KHGX20220322_120125_V06"

#######################################
# We can use the **pyart.io.read_nexrad_archive** module to access our data, passing in the filepath.

radar = pyart.io.read_nexrad_archive(aws_nexrad_level2_file)

#######################################
# Let's take a look at a summary of what fields are available.

list(radar.fields)

#######################################
# Let's plot the reflectivity/velocity fields as a first step to investigating our dataset.
# 
# Note: the reflectivity and velocity fields are in different sweeps, so we will need to specify which sweep to plot in each plot.

fig = plt.figure(figsize=(12,4))
display = pyart.graph.RadarMapDisplay(radar)

ax = plt.subplot(121,
                 projection=ccrs.PlateCarree())

display.plot_ppi_map('reflectivity',
                     sweep=0,
                     ax=ax,
                     colorbar_label='Equivalent Relectivity ($Z_{e}$) \n (dBZ)',
                     vmin=-20,
                     vmax=60)

ax = plt.subplot(122,
                 projection=ccrs.PlateCarree())

display.plot_ppi_map('velocity',
                     sweep=1,
                     ax=ax,
                     colorbar_label='Radial Velocity ($V_{r}$) \n (m/s)',
                     vmin=-70,
                     vmax=70)

#######################################
# Within this plot, we see that the velocity data still has regions that are folded, indicating the dataset has not yet been dealiased.

#######################################%%
# Read NEXRAD Level 3 Data
# ------------------------
# We can also access NEXRAD Level 3 data using Py-ART!
# 
# These datasets have had additional data quality processes applied, including dealiasing.
# 
# Each Level 3 data field is stored in **separate file** - in this example, we will look at the reflectivity and 
# velocity field at the lowest levels. These correspond to the following variable names:
#
# - ``N0U`` - Velocity at the lowest level
# - ``NOQ`` - Reflectivity at the lowest level
# 
# These datasets are also in a different bucket (**unidata-nexrad-level3**), and the files are in a **flat directory structure**
# using the following naming convention:
#
#    ``s3://unidata-nexrad-level3/{radarsite}_{field}_{year}_{month}_{date}_{hour}_{minute}_{second}``
# 
# For example, we can look at data from that same time as the NEXRAD Level 2 data used previously (March 22, 2022 at 1201 UTC)

aws_nexrad_level3_velocity_file = "s3://unidata-nexrad-level3/HGX_N0U_2022_03_22_12_01_25"
aws_nexrad_level3_reflectivity_file = "s3://unidata-nexrad-level3/HGX_N0Q_2022_03_22_12_01_25"

#######################################
#Read our Data using **pyart.io.read_nexrad_level3**

radar_level3_velocity = pyart.io.read_nexrad_level3(aws_nexrad_level3_velocity_file)
radar_level3_reflectivity = pyart.io.read_nexrad_level3(aws_nexrad_level3_reflectivity_file)

#######################################
# Let's confirm that each radar object has a single field:

print('velocity radar object: ', 
      list(radar_level3_velocity.fields),
      'reflectivity radar object: ',
      list(radar_level3_reflectivity.fields)
     )

#######################################
# Plot a Quick Look of our NEXRAD Level 3 Data
# 
# Let's plot the reflectivity/velocity fields as a first step to investigating our dataset.
# 
# Note: the reflectivity and velocity fields are in different radars, so we need to setup different displays.

fig = plt.figure(figsize=(12,4))
reflectivity_display = pyart.graph.RadarMapDisplay(radar_level3_reflectivity)

ax = plt.subplot(121,
                 projection=ccrs.PlateCarree())

reflectivity_display.plot_ppi_map('reflectivity',
                                  ax=ax,
                                  colorbar_label='Equivalent Relectivity ($Z_{e}$) \n (dBZ)',
                                  vmin=-20,
                                  vmax=60)

velocity_display = pyart.graph.RadarMapDisplay(radar_level3_velocity)

ax = plt.subplot(122,
                 projection=ccrs.PlateCarree())

velocity_display.plot_ppi_map('velocity',
                              ax=ax,
                              colorbar_label='Radial Velocity ($V_{r}$) \n (m/s)',
                              vmin=-70,
                              vmax=70)




"""
=========================================
Calculate and Plot hydrometeor classification
=========================================

Calculates a hydrometeor classification and displays the results
"""

# Author: Daniel Wolfensberger (daniel.wolfensberger@meteoswiss.ch)
# License: BSD 3 clause

import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt

import pyart
from open_radar_data import DATASETS

# Read in a sample file
filename = DATASETS.fetch('MLL2217907250U.003.nc')
radar = pyart.io.read_cfradial(filename)

# Read temperature preinterpolated from NWP model
filename = DATASETS.fetch('20220628072500_savevol_COSMO_LOOKUP_TEMP.nc')
nwp_temp = pyart.io.read_cfradial(filename)

# Add temperature to radar object as new field
radar.add_field('temperature', nwp_temp.fields['temperature'])

# Compute attenuation
out = pyart.correct.calculate_attenuation_zphi(
    radar,
    phidp_field='uncorrected_differential_phase',
    temp_field='temperature',
    temp_ref='temperature')
spec_at, pia, cor_z, spec_diff_at, pida, cor_zdr = out
radar.add_field('corrected_reflectivity', cor_z)
radar.add_field('corrected_differential_reflectivity', cor_zdr)
radar.add_field('specific_attenuation', spec_at)

# Compute KDP
kdp, _, _ = pyart.retrieve.kdp_maesaka(
    radar, psidp_field='uncorrected_differential_phase')
radar.add_field('specific_differential_phase', kdp)

# Compute hydrometeor classification
hydro = pyart.retrieve.hydroclass_semisupervised(
    radar,
    refl_field='corrected_reflectivity',
    zdr_field='corrected_differential_reflectivity',
    kdp_field='specific_differential_phase',
    rhv_field='uncorrected_cross_correlation_ratio',
    temp_field='temperature')

radar.add_field('radar_echo_classification', hydro)

# Display hydrometeor classification with categorical colormap
fig, ax = plt.subplots(1,1, figsize=(6, 6))
display = pyart.graph.RadarDisplay(radar)

labels = ['NC','AG', 'CR', 'LR', 'RP', 'RN', 'VI', 'WS', 'MH', 'IH/HDG']
ticks = np.arange(len(labels))
boundaries = np.arange(-0.5, len(labels))
norm = mpl.colors.BoundaryNorm(boundaries, 256)

cax = display.plot_ppi('radar_echo_classification', 0, ax=ax, 
                       norm=norm, ticks=ticks, ticklabs=labels)

ax.set_xlim([-50,50])
ax.set_ylim([-50,50])
ax.set_aspect('equal', 'box')

# For info
# NC = not classified
# AG = aggregates
# CR = ice crystals
# LR = light rain
# RP = rimed particles
# RN = rain
# VI = vertically oriented ice
# WS = wet snow
# MH = melting hail
# IH/HDG = dry hail / high density graupel



"""
=======================================
Convective-Stratiform classification
=======================================
This example shows how to use the updated convective stratiform classifcation algorithm. We show 3 examples,
a summer convective example, an example from Hurricane Ian, and an example from a winter storm.

References
----------
Steiner, M. R., R. A. Houze Jr., and S. E. Yuter, 1995: Climatological
Characterization of Three-Dimensional Storm Structure from Operational
Radar and Rain Gauge Data. J. Appl. Meteor., 34, 1978-2007.
https://doi.org/10.1175/1520-0450(1995)034<1978:CCOTDS>2.0.CO;2.

Yuter, S. E., and R. A. Houze, Jr., 1997: Measurements of raindrop size
distributions over the Pacific warm pool and implications for Z-R relations.
J. Appl. Meteor., 36, 847-867.
https://doi.org/10.1175/1520-0450(1997)036%3C0847:MORSDO%3E2.0.CO;2

Yuter, S. E., R. A. Houze, Jr., E. A. Smith, T. T. Wilheit, and E. Zipser,
2005: Physical characterization of tropical oceanic convection observed in
KWAJEX. J. Appl. Meteor., 44, 385-415. https://doi.org/10.1175/JAM2206.1

Rasmussen, R., M. Dixon, S. Vasiloff, F. Hage, S. Knight, J. Vivekanandan,
and M. Xu, 2003: Snow Nowcasting Using a Real-Time Correlation of Radar
Reflectivity with Snow Gauge Accumulation. J. Appl. Meteorol. Climatol., 42, 20–36.
https://doi.org/10.1175/1520-0450(2003)042%3C0020:SNUART%3E2.0.CO;2
"""

print(__doc__)

# Author: Laura Tomkins (lmtomkin@ncsu.edu)
# License: BSD 3 clause


import pyart
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.colors as mcolors
import cartopy.crs as ccrs

######################################
# **How the algorithm works**
# This first section describes how the convective-stratiform algorithm works (see references for full details). This
# algorithm is a feature detection algorithm and classifies fields as "convective" or "stratiform". The algorithm is
# designed to detect features in a reflectivity field but can also detect features in fields such as rain rate or
# snow rate. In this section we describe the steps of the convective stratiform algorithm and the variables used in
# the function.
# The first step of the algorithm calculates a background average of the field with a circular footprint using a radius
# provided by `bkg_rad_km`. A larger radius will yield a smoother field. The radius needs to be at least double the
# grid spacing, but we recommend at least three times the grid spacing. If using reflectivity, `dB_averaging` should be set
# to True to convert reflectivity to linear Z before averaging, False for rescaled fields such as rain or snow rate.
# `calc_thres` determines the minimum fraction of a circle that is considered in the background average calculation
# (default is 0.75, so the points along the edges where there is less than 75% of a full circle of data,
# the algorithm is not run).
# Once the background average has been calculated, the original field is compared to the background average.  In
# order for points to be considered "convective cores" they must exceed the background value by a certain value or
# simply be greater than the `always_core_thres`. This value is determined by either a cosine scheme, or a scalar
# value (i.e. the reflectivity value must be X times the background value (multiplier; `use_addition=False`),
# or  X greater than the background value (`use_addition=True`) where X is the `scalar_diff`).
# `use_cosine` determines if a cosine scheme or scalar scheme is to be used. If `use_cosine` is True,
# then the `max_diff` and `zero_diff_cos_val` come into use. These values define the cosine scheme that is used  to
# determine the minimum difference between the background and reflectivity value in order for a core to be
# identified. `max_diff` is the maximum difference between the field and the background for a core to be identified,
# or where the cosine function crosses the y-axis. The `zero_diff_cos_val` is where the difference between the  field
# and the background is zero, or where the cosine function crosses the x-axis. Note, if
# `always_core_thres` < `zero_diff_cos_val`, `zero_diff_cos_val` only helps define the shape of the cosine curve and
# all values greater than `always_core_thres` will be considered a convective core. If
# `always_core_thres` > `zero_diff_cos_val` then all values greater than `zero_diff_cos_val` will be considered a
# convective core. We plot some examples of the schemes below:

# plot cosine scheme
pyart.graph.plot_convstrat_scheme(always_core_thres=30, use_cosine=True, max_diff=5, zero_diff_cos_val=45)
# when zero_diff_cos_val is greater than always_core_thres, the difference becomes zero at the zero_diff_cos_val
pyart.graph.plot_convstrat_scheme(always_core_thres=55, use_cosine=True, max_diff=5, zero_diff_cos_val=45)
# alternatively, we can use a simpler scalar difference instead of a cosine scheme
pyart.graph.plot_convstrat_scheme(always_core_thres=40, use_cosine=False, max_diff=None,
                                  zero_diff_cos_val=None, use_addition=True, scalar_diff=2)
# if you are interested in picking up weak features, you can also use the scalar difference as a multiplier instead,
# so very weak features do not have to be that different from the background to be classified as convective.
pyart.graph.plot_convstrat_scheme(always_core_thres=40, use_cosine=False, max_diff=None,
                                  zero_diff_cos_val=None, use_addition=False, scalar_diff=2)

######################################
# Once the cores are identified, there is an option to remove speckles (`remove_small_objects`) smaller than a  given
# size (`min_km2_size`).
# After the convective cores are identified, We then incorporate convective radii using
# `val_for_max_conv_rad` and `max_conv_rad_km`. The convective radii act as a dilation and are used to classify
# additional points around the cores as convective that may not have been identified previously.  The
# `val_for_max_conv_rad` is the value where the maximum convective radius is applied and the `max_conv_rad_km` is the
# maximum convective radius. Values less than the `val_for_max_conv_rad` are assigned a convective radius using a step
# function.
# Finally, the points are classified as NOSFCECHO (threshold set with `min_dBZ_used`; 0), WEAKECHO (threshold set with
# `weak_echo_thres`; 3), SF (stratiform; 1), CONV (convective; 2).

######################################
# **Classification of summer convective example**
# Our first example classifies echo from a summer convective event. We use a cosine scheme to classify the convective
# points.

# Now let's do a classification with our parameters
# read in file
filename = pyart.testing.get_test_data('swx_20120520_0641.nc')
radar = pyart.io.read(filename)

# extract the lowest sweep
radar = radar.extract_sweeps([0])

# interpolate to grid
grid = pyart.map.grid_from_radars(
    (radar,), grid_shape=(1, 201, 201),
    grid_limits=((0, 10000), (-50000.0, 50000.0), (-50000.0, 50000.0)),
    fields=['reflectivity_horizontal'])

# get dx dy
dx = grid.x['data'][1] - grid.x['data'][0]
dy = grid.y['data'][1] - grid.y['data'][0]

# convective stratiform classification
convsf_dict = pyart.retrieve.conv_strat_yuter(grid, dx, dy, refl_field='reflectivity_horizontal', always_core_thres=40,
                                              bkg_rad_km=20, use_cosine=True, max_diff=5, zero_diff_cos_val=55,
                                              weak_echo_thres=10, max_conv_rad_km=2)

# add to grid object
# mask zero values (no surface echo)
convsf_masked = np.ma.masked_equal(convsf_dict['convsf']['data'], 0)
# mask 3 values (weak echo)
convsf_masked = np.ma.masked_equal(convsf_masked, 3)
# add dimension to array to add to grid object
convsf_dict['convsf']['data'] = convsf_masked[None,:,:]
# add field
grid.add_field('convsf', convsf_dict['convsf'], replace_existing=True)

# create plot using GridMapDisplay
# plot variables
display = pyart.graph.GridMapDisplay(grid)
magma_r_cmap = plt.get_cmap('magma_r')
ref_cmap = mcolors.LinearSegmentedColormap.from_list('ref_cmap', magma_r_cmap(np.linspace(0, 0.9, magma_r_cmap.N)))
projection = ccrs.LambertConformal(central_latitude=radar.latitude['data'][0],
                                   central_longitude=radar.longitude['data'][0])
######################################
# Note how the convective stratiform field has less data around the edges compared to the reflectivity
# field. This is because the function is designed to only compute the background radius where the footprint has 75%
# of data, so along the edges there is not enough data to fulfill this requirement. The footprint percentage can be
# changed using the variable `calc_thres`.

# plot
plt.figure(figsize=(10,4))
ax1=plt.subplot(1,2,1, projection=projection)
display.plot_grid('reflectivity_horizontal', vmin=5, vmax=45, cmap=ref_cmap, projection=projection,
                  transform=ccrs.PlateCarree(), ax=ax1)
ax2=plt.subplot(1,2,2, projection=projection)
display.plot_grid('convsf', vmin=0, vmax=2, cmap=plt.get_cmap('viridis', 3), projection=projection, ax=ax2,
                  transform=ccrs.PlateCarree(), ticks=[1/3, 1, 5/3], ticklabs=['', 'Stratiform', 'Convective'])
plt.show()


######################################
# In addition to the default convective-stratiform classification, the function also returns an underestimate
# (convsf_under) and an overestimate (convsf_over) to take into consideration the uncertainty when choosing
# classification parameters. The under and overestimate use the same parameters, but vary the input field by a
# certain value (default is 5 dBZ, can be changed with `estimate_offset`). The estimation can be turned off (
# `estimate_flag=False`), but we recommend keeping it turned on.

# mask weak echo and no surface echo
convsf_masked = np.ma.masked_equal(convsf_dict['convsf']['data'], 0)
convsf_masked = np.ma.masked_equal(convsf_masked, 3)
convsf_dict['convsf']['data'] = convsf_masked
# underest.
convsf_masked = np.ma.masked_equal(convsf_dict['convsf_under']['data'], 0)
convsf_masked = np.ma.masked_equal(convsf_masked, 3)
convsf_dict['convsf_under']['data'] = convsf_masked
# overest.
convsf_masked = np.ma.masked_equal(convsf_dict['convsf_over']['data'], 0)
convsf_masked = np.ma.masked_equal(convsf_masked, 3)
convsf_dict['convsf_over']['data'] = convsf_masked

# Plot each estimation
plt.figure(figsize=(10,4))
ax1=plt.subplot(131)
ax1.pcolormesh(convsf_dict['convsf']['data'][0,:,:], vmin=0, vmax=2, cmap=plt.get_cmap('viridis', 3))
ax1.set_title('Best estimate')
ax1.set_aspect('equal')
ax2=plt.subplot(132)
ax2.pcolormesh(convsf_dict['convsf_under']['data'], vmin=0, vmax=2, cmap=plt.get_cmap('viridis', 3))
ax2.set_title('Underestimate')
ax2.set_aspect('equal')
ax3=plt.subplot(133)
ax3.pcolormesh(convsf_dict['convsf_over']['data'], vmin=0, vmax=2, cmap=plt.get_cmap('viridis', 3))
ax3.set_title('Overestimate')
ax3.set_aspect('equal')
plt.show()

######################################
# **Tropical example**
# Let's get a NEXRAD file from Hurricane Ian

# Read in file
nexrad_file = 's3://noaa-nexrad-level2/2022/09/28/KTBW/KTBW20220928_190142_V06'
radar = pyart.io.read_nexrad_archive(nexrad_file)

# extract the lowest sweep
radar = radar.extract_sweeps([0])

# interpolate to grid
grid = pyart.map.grid_from_radars(
    (radar,), grid_shape=(1, 201, 201),
    grid_limits=((0, 10000), (-200000.0, 200000.0), (-200000.0, 200000.0)),
    fields=['reflectivity'])

# get dx dy
dx = grid.x['data'][1] - grid.x['data'][0]
dy = grid.y['data'][1] - grid.y['data'][0]

# convective stratiform classification
convsf_dict = pyart.retrieve.conv_strat_yuter(grid, dx, dy, refl_field='reflectivity', always_core_thres=40,
                                              bkg_rad_km=20, use_cosine=True, max_diff=3, zero_diff_cos_val=55,
                                              weak_echo_thres=5, max_conv_rad_km=2, estimate_flag=False)

# add to grid object
# mask zero values (no surface echo)
convsf_masked = np.ma.masked_equal(convsf_dict['convsf']['data'], 0)
# mask 3 values (weak echo)
convsf_masked = np.ma.masked_equal(convsf_masked, 3)
# add dimension to array to add to grid object
convsf_dict['convsf']['data'] = convsf_masked[None,:,:]
# add field
grid.add_field('convsf', convsf_dict['convsf'], replace_existing=True)

# create plot using GridMapDisplay
# plot variables
display = pyart.graph.GridMapDisplay(grid)
magma_r_cmap = plt.get_cmap('magma_r')
ref_cmap = mcolors.LinearSegmentedColormap.from_list('ref_cmap', magma_r_cmap(np.linspace(0, 0.9, magma_r_cmap.N)))
projection = ccrs.LambertConformal(central_latitude=radar.latitude['data'][0],
                                   central_longitude=radar.longitude['data'][0])
# plot
plt.figure(figsize=(10,4))
ax1=plt.subplot(1,2,1, projection=projection)
display.plot_grid('reflectivity', vmin=5, vmax=45, cmap=ref_cmap, projection=projection,
                  transform=ccrs.PlateCarree(), ax=ax1, axislabels_flag=False)
ax2=plt.subplot(1,2,2, projection=projection)
display.plot_grid('convsf', vmin=0, vmax=2, cmap=plt.get_cmap('viridis', 3), projection=projection,
                  axislabels_flag=False, transform=ccrs.PlateCarree(), ticks=[1/3, 1, 5/3],
                  ticklabs=['', 'Stratiform', 'Convective'], ax=ax2)
plt.show()

######################################
# **Winter storm example with image muting**
# Here is a final example of the convective stratiform classification using an example from a winter storm. Before
# doing the classification, we image mute the reflectivity to remove regions with melting or mixed precipitation. We
# then rescale the reflectivity to snow rate (Rasumussen et al. 2003). We recommend using a rescaled reflectivity
# to do the classification, but if you do make sure to changed dB_averaging to False because this parameter is  used
# to convert reflectivity to a linear value before averaging (set dB_averaging to True for reflectivity fields in
# dBZ units).
# In this example, note how we change some of the other parameters since we are classifying snow rate instead of
# reflecitivity.

# Read in file
nexrad_file = 's3://noaa-nexrad-level2/2021/02/07/KOKX/KOKX20210207_161413_V06'
radar = pyart.io.read_nexrad_archive(nexrad_file)

# extract the lowest sweep
radar = radar.extract_sweeps([0])

# interpolate to grid
grid = pyart.map.grid_from_radars(
    (radar,), grid_shape=(1, 201, 201),
    grid_limits=((0, 10000), (-200000.0, 200000.0), (-200000.0, 200000.0)),
    fields=['reflectivity', 'cross_correlation_ratio'])

# image mute grid object
grid = pyart.util.image_mute_radar(grid, 'reflectivity', 'cross_correlation_ratio', 0.97, 20)

# convect non-muted reflectivity to snow rate
nonmuted_ref = grid.fields['nonmuted_reflectivity']['data'][0,:,:]
nonmuted_ref = np.ma.masked_invalid(nonmuted_ref)

nonmuted_ref_linear = 10 ** (nonmuted_ref / 10)  # mm6/m3
snow_rate = (nonmuted_ref_linear/57.3)**(1/1.67) #

# add to grid
snow_rate_dict = {
        'data': snow_rate[None,:,:],
        'standard_name': 'snow_rate',
        'long_name': 'Snow rate converted from linear reflectivity',
        'units': 'mm/hr',
        'valid_min': 0,
        'valid_max': 40500}
grid.add_field('snow_rate', snow_rate_dict, replace_existing=True)

# get dx dy
dx = grid.x['data'][1] - grid.x['data'][0]
dy = grid.y['data'][1] - grid.y['data'][0]

# convective stratiform classification
convsf_dict = pyart.retrieve.conv_strat_yuter(grid, dx, dy, refl_field='snow_rate', dB_averaging=False,
                                            always_core_thres=4,bkg_rad_km=40, use_cosine=True, max_diff=1.5,
                                            zero_diff_cos_val=5,weak_echo_thres=0, min_dBZ_used=0,
                                            max_conv_rad_km=1, estimate_flag=False)

# add to grid object
# mask zero values (no surface echo)
convsf_masked = np.ma.masked_equal(convsf_dict['convsf']['data'], 0)
# mask 3 values (weak echo)
convsf_masked = np.ma.masked_equal(convsf_masked, 3)
# add dimension to array to add to grid object
convsf_dict['convsf']['data'] = convsf_masked[None,:,:]
# add field
grid.add_field('convsf', convsf_dict['convsf'], replace_existing=True)

# create plot using GridMapDisplay
# plot variables
display = pyart.graph.GridMapDisplay(grid)
magma_r_cmap = plt.get_cmap('magma_r')
ref_cmap = mcolors.LinearSegmentedColormap.from_list('ref_cmap', magma_r_cmap(np.linspace(0, 0.9, magma_r_cmap.N)))
projection = ccrs.LambertConformal(central_latitude=radar.latitude['data'][0],
                                   central_longitude=radar.longitude['data'][0])
# plot
plt.figure(figsize=(10,4))
ax1=plt.subplot(1,2,1, projection=projection)
display.plot_grid('snow_rate', vmin=0, vmax=10, cmap=plt.get_cmap('viridis'), projection=projection,
                  transform=ccrs.PlateCarree(), ax=ax1, axislabels_flag=False)
ax2=plt.subplot(1,2,2, projection=projection)
display.plot_grid('convsf', vmin=0, vmax=2, cmap=plt.get_cmap('viridis', 3), projection=projection,
                  axislabels_flag=False, transform=ccrs.PlateCarree(), ticks=[1/3, 1, 5/3],
                  ticklabs=['', 'Stratiform', 'Convective'], ax=ax2)
plt.show()

######################################
# **Summary of recommendations and best practices**
# - Tune your parameters to your specific purpose
# - Use a rescaled field if possible (i.e. linear reflectivity, rain or snow rate)
# - Keep `estimate_flag=True` to see uncertainty in classification


"""
======================================
Compare Two Radars Using Gatemapper
======================================

Map the reflectivity field of a single radar in Antenna coordinates to
another radar in Antenna coordinates and compare the fields.

"""
print(__doc__)

# Author: Max Grover (mgrover@anl.gov) and Bobby Jackson (rjackson@anl.gov)
# License: BSD 3 clause

import cartopy.crs as ccrs
import matplotlib.pyplot as plt
import numpy as np
import warnings

import pyart
from pyart.testing import get_test_data
warnings.filterwarnings("ignore")

######################################
# **Read in the Data**
#
# For this example, we use two XSAPR radars from our test data.

# read in the data from both XSAPR radars
xsapr_sw_file = get_test_data('swx_20120520_0641.nc')
xsapr_se_file = get_test_data('sex_20120520_0641.nc')
radar_sw = pyart.io.read_cfradial(xsapr_sw_file)
radar_se = pyart.io.read_cfradial(xsapr_se_file)

######################################
# **Filter and Configure the GateMapper**
#
# We are interested in mapping the southwestern radar to the
# southeastern radar. Before running our gatemapper, we add a
# filter for only positive reflectivity values.
# We also need to set a distance (meters) and time (seconds)
# between the source and destination gate allowed for an
# adequate match), using the distance_tolerance/time_tolerance variables.

gatefilter = pyart.filters.GateFilter(radar_sw)
gatefilter.exclude_below('reflectivity_horizontal', 20)
gmapper = pyart.map.GateMapper(radar_sw,
                               radar_se,
                               distance_tolerance=500.,
                               time_tolerance=60,
                               gatefilter_src=gatefilter)
radar_sw_mapped_to_radar_se = gmapper.mapped_radar(['reflectivity_horizontal'])

######################################
# **Plot the Original Data**
#
# Let's take a look at our original fields - notice the difference
# in reflectivity values!

fig = plt.figure(figsize=(16, 6))
ax = plt.subplot(121, projection=ccrs.PlateCarree())

# Plot the southwestern radar
disp1 = pyart.graph.RadarMapDisplay(radar_sw)
disp1.plot_ppi_map('reflectivity_horizontal',
                   sweep=1,
                   ax=ax,
                   vmin=-20,
                   vmax=70,
                   min_lat=36,
                   max_lat=37,
                   min_lon=-98,
                   max_lon=-97,
                   lat_lines=np.arange(36, 37.25, .25),
                   lon_lines=np.arange(-98, -96.75, .25))

ax2 = plt.subplot(122, projection=ccrs.PlateCarree())
disp2 = pyart.graph.RadarMapDisplay(radar_se)
disp2.plot_ppi_map('reflectivity_horizontal',
                   sweep=1,
                   ax=ax2,
                   vmin=-20,
                   vmax=70,
                   min_lat=36,
                   max_lat=37,
                   min_lon=-98,
                   max_lon=-97,
                   lat_lines=np.arange(36, 37.25, .25),
                   lon_lines=np.arange(-98, -96.75, .25))

######################################
# Now, we can compare our original field from the southwestern radar,
# to the new remapped field - there are similarities...

fig = plt.figure(figsize=(16, 6))
ax = plt.subplot(121, projection=ccrs.PlateCarree())

# Plot the southeastern radar
disp1 = pyart.graph.RadarMapDisplay(radar_se)
disp1.plot_ppi_map('reflectivity_horizontal',
                   sweep=1,
                   ax=ax,
                   vmin=-20,
                   vmax=70,
                   min_lat=36,
                   max_lat=37,
                   min_lon=-98,
                   max_lon=-97,
                   lat_lines=np.arange(36, 37.25, .25),
                   lon_lines=np.arange(-98, -96.75, .25))

# Plot the southwestern radar mapped to the southeastern radar
ax2 = plt.subplot(122, projection=ccrs.PlateCarree())
disp2 = pyart.graph.RadarMapDisplay(radar_sw_mapped_to_radar_se)
disp2.plot_ppi_map('reflectivity_horizontal',
                   sweep=1,
                   ax=ax2,
                   vmin=-20,
                   vmax=70,
                   min_lat=36,
                   max_lat=37,
                   min_lon=-98,
                   max_lon=-97,
                   lat_lines=np.arange(36, 37.25, .25),
                   lon_lines=np.arange(-98, -96.75, .25))

######################################
# **Calculate and Plot the Difference**
#
# It can be difficult to "eyeball" the difference between these two fields.
# Fortunately, now that our radars match coordinates, we can plot a difference.
# Keep in mind there is a time difference of ~ 1 minute between these plots,
# leading to small difference due to the precipitation moving through the 
# domain over the course of that minute.

# Extract the numpy arrays for our reflectivity fields
reflectivity_se_radar = radar_se.fields['reflectivity_horizontal']['data']
reflectivity_sw_radar = radar_sw_mapped_to_radar_se.fields['reflectivity_horizontal']['data']

# Calculate the difference between the southeastern and southwestern radar
reflectivity_difference = reflectivity_se_radar - reflectivity_sw_radar

# Add a field like this to the radar_se radar object
radar_se.add_field_like('reflectivity_horizontal',
                        field_name='reflectivity_bias',
                        data=reflectivity_difference)

# Setup our figure
fig = plt.figure(figsize=(8, 6))
ax = plt.subplot(111, projection=ccrs.PlateCarree())

# Plot the difference field
disp1 = pyart.graph.RadarMapDisplay(radar_se)
disp1.plot_ppi_map('reflectivity_bias',
                   cmap='pyart_balance',
                   title='Reflectivity Difference \n XSAPR Southwest - XSPAR Southeast',
                   sweep=1,
                   ax=ax,
                   vmin=-30,
                   vmax=30,
                   min_lat=36,
                   max_lat=37,
                   min_lon=-98,
                   max_lon=-97,
                   lat_lines=np.arange(36, 37.25, .25),
                   lon_lines=np.arange(-98, -96.75, .25))


######################################
# **Plot a Histogram for Comparison**
#
# Another way of plotting the comparison here is using
# a 2-dimensional histogram,which is more helpful in this
# case where our scans don't neccessarily match exactly in time.

# Include elevations above the lowest one
incl_gates = np.argwhere(radar_sw_mapped_to_radar_se.elevation['data'] > 1.)

# Filter the reflectivity fields using the filter created above
refl_se = reflectivity_se_radar[incl_gates, :]
refl_sw = reflectivity_sw_radar[incl_gates, :]

# Make sure not include masked values
values_without_mask = np.logical_and(~refl_se.mask, ~refl_sw.mask)
refl_se = refl_se[values_without_mask]
refl_sw = refl_sw[values_without_mask]

# Set the bins for our histogram
bins = np.arange(-10, 60, 1)

# Create the 2D histogram using the flattened numpy arrays
hist = np.histogram2d(refl_se.flatten(), refl_sw.flatten(), bins=bins)[0]
hist = np.ma.masked_where(hist == 0, hist)

# Setup our figure
fig = plt.figure(figsize=(8, 6))

# Create a 1-1 comparison
x, y = np.meshgrid((bins[:-1] + bins[1:])/2., (bins[:-1] + bins[1:])/2.)
c = plt.pcolormesh(x, y, np.log10(hist.T), cmap='pyart_HomeyerRainbow')

# Add a colorbar and labels
plt.colorbar(c, label='$log_{10}$ counts')
plt.xlabel('XSAPR Southeast $Z_{H}$ [dBZ]')
plt.ylabel('XSAPR Southwest $Z_{H}$ [dBZ]')

plt.show()



""" Unit Tests for Py-ART's retrieve/echo_class.py module. """

import numpy as np
from numpy.testing import assert_allclose
import pytest

import pyart


def test_steiner_conv_strat_default():
    grid = pyart.testing.make_storm_grid()
    eclass_default = pyart.retrieve.steiner_conv_strat(grid)
    assert np.all(eclass_default['data'][25] == np.array(
        [0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
         2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0]))

@pytest.mark.parametrize('area_relation', ['small', 'medium', 'large', 'sgp'])
def test_steiner_conv_strat_modify_area(area_relation):
    grid = pyart.testing.make_storm_grid()
    eclass = pyart.retrieve.steiner_conv_strat(grid,
                                               area_relation=area_relation)
    assert eclass['data'].min() == 0
    assert eclass['data'].max() == 2

def test_conv_strat__yuter_default():
    grid = pyart.testing.make_storm_grid()
    dict = pyart.retrieve.conv_strat_yuter(grid, bkg_rad_km=50)

    assert 'convsf' in dict.keys()
    assert 'convsf_under' in dict.keys()
    assert 'convsf_over' in dict.keys()
    assert np.all(dict['convsf']['data'][25] == np.array(
        [0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
         2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0]))

def test_conv_strat_yuter_noest():
    grid = pyart.testing.make_storm_grid()
    dict = pyart.retrieve.conv_strat_yuter(grid, bkg_rad_km=50, estimate_flag=False)

    assert 'convsf' in dict.keys()
    assert 'convsf_under' not in dict.keys()
    assert 'convsf_over' not in dict.keys()

def test_hydroclass_semisupervised():
    radar = pyart.io.read(pyart.testing.NEXRAD_ARCHIVE_MSG31_FILE)
    temp_dict = pyart.config.get_metadata('temperature')
    temp = np.empty(radar.fields['reflectivity']['data'].shape)
    temp[:] = -10.0
    temp_dict['data'] = temp
    radar.add_field('temperature', temp_dict, replace_existing=True)
    radar.fields['specific_differential_phase'] = radar.fields.pop('differential_phase')
    radar_small = radar.extract_sweeps([0])

    mass_centers = pyart.retrieve.echo_class._get_mass_centers(10e9)

    hydro_nofreq = pyart.retrieve.hydroclass_semisupervised(radar_small)
    assert 'units' in hydro_nofreq.keys()
    assert 'standard_name' in hydro_nofreq.keys()
    assert 'long_name' in hydro_nofreq.keys()
    assert 'coordinates' in hydro_nofreq.keys()

    assert hydro_nofreq['data'].dtype == 'int64'
    assert_allclose(
        hydro_nofreq['data'][0][0:5],
        [6, 6, 6, 6, 6])
    assert_allclose(
        hydro_nofreq['data'][0][-5:],
        [2, 2, 2, 2, 2])

    radar_small.instrument_parameters['frequency'] = {'data': np.array([5e9])}
    hydro_freq = pyart.retrieve.hydroclass_semisupervised(radar_small)
    assert_allclose(
        hydro_freq['data'][0][0:5],
        [6, 6, 6, 6, 6])
    assert_allclose(
        hydro_freq['data'][0][-5:],
        [2, 2, 2, 2, 2])

    hydro = pyart.retrieve.hydroclass_semisupervised(
        radar_small, mass_centers=mass_centers)
    assert_allclose(
        hydro['data'][0][0:5],
        [6, 6, 6, 6, 6])
    assert_allclose(
        hydro['data'][0][-5:],
       [2, 2, 2, 2, 2])


def test_data_limits_table():
    dlimits_dict = pyart.retrieve.echo_class._data_limits_table()
    test_dict = {'Zh': (60.0, -10.0),
                 'ZDR': (5.0, -5.0),
                 'KDP': (7.0, -10.0),
                 'RhoHV': (-5.23, -50.0)}

    assert type(dlimits_dict) == dict
    assert dlimits_dict == test_dict


def test_get_freq():
    freq_band_s = pyart.retrieve.get_freq_band(3e9)
    assert freq_band_s == 'S'

    freq_band_c = pyart.retrieve.get_freq_band(6e9)
    assert freq_band_c == 'C'

    freq_band_x = pyart.retrieve.get_freq_band(10e9)
    assert freq_band_x == 'X'

    freq_band_bad = pyart.retrieve.get_freq_band(10)
    assert freq_band_bad == None


def test_mass_centers_dict():
    mass_centers_dict =  pyart.retrieve.echo_class._mass_centers_table()

    assert type(mass_centers_dict)
    assert 'C' in mass_centers_dict.keys()
    assert 'X' in mass_centers_dict.keys()

    assert_allclose(
        mass_centers_dict['C'][0],
        [1.35829e+01,  4.06300e-01,  4.97000e-02,  9.86800e-01,
         1.33030e+03], atol=1e-7)
    assert_allclose(
        mass_centers_dict['C'][8],
        [5.06186e+01, -6.49000e-02,  9.46000e-02,  9.90400e-01,
         1.17990e+03], atol=1e-7)

    assert_allclose(
        mass_centers_dict['X'][0],
        [1.9077e+01, 4.1390e-01, 9.9000e-03, 9.8410e-01,
         1.0617e+03], atol=1e-7)

    assert_allclose(
        mass_centers_dict['X'][8],
        [4.42216e+01, -3.41900e-01,  6.87000e-02,  9.68300e-01,
         1.27270e+03], atol=1e-7)


def test_mass_centers():
    mass_centers_good_freq = pyart.retrieve.echo_class._get_mass_centers(10e9)
    assert_allclose(
        mass_centers_good_freq[0],
        [1.90770e+01,  4.13900e-01,  9.90000e-03,  9.84100e-01,
         1.06170e+03], atol=1e-7)

    low_bad_freq = pyart.retrieve.echo_class._get_mass_centers(3e9)
    assert_allclose(
        low_bad_freq[0],
        [1.35829e+01,  4.06300e-01,  4.97000e-02,  9.86800e-01,
         1.33030e+03], atol=1e-7)

    high_bad_freq = pyart.retrieve.echo_class._get_mass_centers(13e9)
    assert_allclose(
        high_bad_freq[0],
        [1.90770e+01,  4.13900e-01,  9.90000e-03,  9.84100e-01,
         1.06170e+03], atol=1e-7)


def test_assign_to_class():
    rhohv = np.array(([0.2, 0.2, 0.2, 0.2, 0.2],
                      [0.2, 0.2, 0.2, 0.2, 0.2]))
    zdr = np.array(([-7.8, -7.8, -7.8, -7.8, -7.8],
                    [-7.8, -7.8, -7.8, -7.8, -7.8]))
    kdp = np.array(([180.5, 180.5, 180.5, 180.5, 180.5],
                    [180.5, 180.5, 180.5, 180.5, 180.5]))
    relh = np.array(([10.0, 10.0, 10.0, 10.0, 10.0],
                     [10.0, 10.0, 10.0, 10.0, 10.0]))
    zh = np.array(([-32.0, -32.0, -32.0, -32.0, -32.0],
                   [-32.0, -32.0, -32.0, -32.0, -32.0]))

    mass_centers = pyart.retrieve.echo_class._get_mass_centers(10e9)
    hydroclass, min_dist = pyart.retrieve.echo_class._assign_to_class(
        zh, zdr, kdp, rhohv, relh, mass_centers)

    assert_allclose(hydroclass[0], [7, 7, 7, 7, 7], atol=1e-7)
    assert_allclose(
        min_dist[0],
        [227.0343910, 227.0343910, 227.0343910,
         227.0343910, 227.0343910], atol=1e-7)


def test_standardize():
    kdp = np.array(([180.5, 180.5, 180.5, 180.5, 180.5],
                    [180.5, 180.5, 180.5, 180.5, 180.5]))
    field_std_kdp = pyart.retrieve.echo_class._standardize(
        kdp, 'KDP', mx=180.0, mn=181.0)
    assert_allclose(field_std_kdp[0],
                    [-1.0, -1.0, -1.0, -1.0, -1.0],
                    atol=1e-6)

    relh = np.array(([10.0, 10.0, 10.0, 10.0, 10.0],
                     [10.0, 10.0, 10.0, 10.0, 10.0]))
    field_std_relh = pyart.retrieve.echo_class._standardize(
        relh, 'relH', mx=9.0, mn=5.0)
    assert_allclose(
        field_std_relh[0],
        [0.024994, 0.024994, 0.024994, 0.024994, 0.024994],
        atol=1e-6)

    rhohv = np.array(([0.2, 0.2, 0.2, 0.2, 0.2],
                      [0.2, 0.2, 0.2, 0.2, 0.2]))
    field_std_rhohv = pyart.retrieve.echo_class._standardize(
        rhohv, 'RhoHV', mx=0.1, mn=1.0)
    assert_allclose(
        field_std_rhohv[0],
        [-1.0, -1.0, -1.0, -1.0, -1.0],
        atol=1e-6)

    field_std_no_limits = pyart.retrieve.echo_class._standardize(
        rhohv, 'RhoHV', mx=None, mn=None)
    assert_allclose(
        field_std_no_limits[0],
        [1.0, 1.0, 1.0, 1.0, 1.0],
        atol=1e-6)

    pytest.raises(
        ValueError, pyart.retrieve.echo_class._standardize, rhohv, 'foo')



"""
===========================
Create a two panel RHI plot
===========================

An example which creates a two panel RHI plot of a cfradial file.  The fields
included in the two panels are reflectivity and doppler velocity.

"""
print(__doc__)

# Author: Max Grover (mgrover@anl.gov)
# License: BSD 3 clause

import matplotlib.pyplot as plt
import pyart
from pyart.testing import get_test_data
import numpy as np


# Read the data and create the display object
filename = get_test_data('sgpxsaprrhicmacI5.c0.20110524.015604_NC4.nc')
radar = pyart.io.read_cfradial(filename)
display = pyart.graph.RadarDisplay(radar)

# Fields to plot and ranges
fields_to_plot = ['reflectivity_horizontal', 'mean_doppler_velocity']
ranges = [(-20, 20), (-17.0, 17.0)]
cmaps = ['pyart_HomeyerRainbow', 'pyart_balance']

# Plot the data
nplots = len(fields_to_plot)
plt.figure(figsize=[5 * nplots, 4])

# Plot each field
for plot_num in range(nplots):
    field = fields_to_plot[plot_num]
    vmin, vmax = ranges[plot_num]
    cmap = cmaps[plot_num]

    plt.subplot(1, nplots, plot_num + 1)
    display.plot(field, 0, vmin=vmin, vmax=vmax, title_flag=False, cmap=cmap)
    display.set_limits(ylim=[0, 17])

# Grab the fixed angle and time from the first sweep
fixed_angle = radar.fixed_angle['data'][0]
time = radar.time["units"][13:]

# Add the metadata to the title
plt.suptitle(f"Reflectivity and Velocity \n Azimuth: {np.around(fixed_angle, 3)}\u00B0 {time} UTC")
plt.show()



"""
Calculate the composite reflectivity 

"""

import copy

import numpy as np
from netCDF4 import num2date
from pandas import to_datetime
from scipy.interpolate import interp2d

from pyart.core import Radar


def composite_reflectivity(radar, field='reflectivity', gatefilter=None):

    """
    Composite Reflectivity 
    
    Often a legacy product, composite reflectivity is: 
    "A display or mapping of the maximum radar reflectivity factor at any
    altitude as a function of position on the ground." - AMS Glossary
    This is more useful for the dry regions of the world, where maximum
    reflectivity values are found aloft, as opposed to the lowest scan.
    Alternatively this is useful for comparing to NWP since composite Z
    is a standard output of NWP. 

    Why this is not as easy as one would think: Turns out the data are
    not natively stored with index 0 being azimuth 0. Likely due to the
    physical spinning of the radar antenna. 

    Author: Randy J. Chase (@dopplerchase) 

    Parameters
    ----------
    radar : Radar
        Radar object used.
    field : str
        Reflectivity field name to use to look up reflectivity data. In the
        radar object. Default field name is 'reflectivity'.
    gatefilter : GateFilter
        GateFilter instance. None will result in no gatefilter mask being
        applied to data. 

    Returns
    -------
    radar : Radar
        The radar object containing the radar dimensions, metadata and
        composite field.

    References
    ----------
    American Meteorological Society, 2022: "Composite reflectivity". Glossary of Meteorology, 
    http://glossary.ametsoc.org/wiki/Composite_reflectivity

    """

    # Determine the lowest sweep (used for metadata and such)
    minimum_sweep = np.min(radar.sweep_number["data"])

    # loop over all measured sweeps 
    for sweep in sorted(radar.sweep_number['data']):
        # get start and stop index numbers
        sweep_slice = radar.get_slice(sweep)

        # grab radar data 
        z = radar.get_field(sweep, field)
        z_dtype = z.dtype

        # Use gatefilter
        if gatefilter is not None:
            mask_sweep = gatefilter.gate_excluded[sweep_slice, :]
            z = np.ma.masked_array(z, mask_sweep)

        # extract lat lons 
        lon = radar.gate_longitude['data'][sweep_slice, :]
        lat = radar.gate_latitude['data'][sweep_slice, :]

        # get the range and time
        ranges = radar.range["data"]
        time = radar.time["data"]

        # get azimuth
        az = radar.azimuth['data'][sweep_slice]
        # get order of azimuths 
        az_ids = np.argsort(az)

        # reorder azs so they are in order 
        az = az[az_ids]
        z = z[az_ids]
        lon = lon[az_ids]
        lat = lat[az_ids]
        time = time[az_ids]

        # if the first sweep, store re-ordered lons/lats
        if sweep == minimum_sweep:
            azimuth_final = az
            time_final = time
            lon_0 = copy.deepcopy(lon)
            lon_0[-1, :] = lon_0[0, :]
            lat_0 = copy.deepcopy(lat)
            lat_0[-1, :] = lat_0[0, :]

        else:
            # Configure the intperpolator
            z_interpolator = interp2d(ranges,
                                      az,
                                      z,
                                      kind='linear')

            # Apply the interpolation
            z = z_interpolator(ranges, azimuth_final)

        # if first sweep, create new dim, otherwise concat them up 
        if sweep == minimum_sweep:
            z_stack = copy.deepcopy(z[np.newaxis, :, :])
            az_final = radar.azimuth['data']
        else:
            z_stack = np.concatenate([z_stack, z[np.newaxis, :, :]])

    # now that the stack is made, take max across vertical 
    compz = z_stack.max(axis=0).astype(z_dtype)

    # since we are using the whole volume scan, report mean time 
    dtime = to_datetime(num2date(radar.time['data'],
                        radar.time['units']).astype(str))
    dtime = dtime.mean()

    # return dict, because this is was pyart does with lots of things 
    fields = {}
    fields['composite_reflectivity'] = {
        'data': compz, 'units': 'dBZ',
        'long_name': 'composite_reflectivity',
        'comment': 'composite reflectivity computed from calculating the max radar value in each radar gate vertically after reordering'}

    time = radar.time.copy()
    time['data'] = time_final
    time['mean'] = dtime

    gate_longitude = radar.gate_longitude.copy()
    gate_longitude['data'] = lon_0
    gate_longitude['comment'] = 'reordered longitude grid, [az,range]'
    gate_latitude = radar.gate_latitude.copy()
    gate_latitude['data'] = lat_0
    gate_latitude['comment'] = 'reordered latitude grid, [az,range]'


    _range = radar.range.copy()
    metadata = radar.metadata.copy()
    scan_type = radar.scan_type
    latitude = radar.latitude.copy()
    longitude = radar.longitude.copy()
    altitude = radar.altitude.copy()
    instrument_parameters = radar.instrument_parameters

    sweep_number = radar.sweep_number.copy()
    sweep_number['data'] = np.array([0], dtype='int32')
    sweep_mode = radar.sweep_mode.copy()
    sweep_mode['data'] = np.array([radar.sweep_mode['data'][0]])
    ray_shape = compz.shape[0]
    azimuth = radar.azimuth.copy()
    azimuth['data'] = azimuth_final
    elevation = radar.elevation.copy()
    elevation['data'] = np.zeros(ray_shape, dtype='float32')
    fixed_angle = radar.fixed_angle.copy()
    fixed_angle['data'] = np.array([0.0], dtype='float32')
    sweep_start_ray_index = radar.sweep_start_ray_index.copy()
    sweep_start_ray_index['data'] = np.array([0], dtype='int32')
    sweep_end_ray_index = radar.sweep_end_ray_index.copy()
    sweep_end_ray_index['data'] = np.array([ray_shape-1], dtype='int32')

    return Radar(
        time, _range, fields, metadata, scan_type,
        latitude, longitude, altitude,
        sweep_number, sweep_mode, fixed_angle, sweep_start_ray_index,
        sweep_end_ray_index,
        azimuth, elevation,
        instrument_parameters=instrument_parameters)



"""
======================================================
Create a multiple panel RHI plot from a CF/Radial file
======================================================

An example which creates a RHI plot of a CF/Radial file using
a RadarDisplay object.

"""
print(__doc__)

import matplotlib.pyplot as plt
import netCDF4
import pyart
from pyart.testing import get_test_data

filename = get_test_data('sgpxsaprrhicmacI5.c0.20110524.015604_NC4.nc')

# create the plot using RadarDisplay
radar = pyart.io.read_cfradial(filename)
radar.metadata['instrument_name'] = 'XSARP'
display = pyart.graph.RadarDisplay(radar)

fig = plt.figure(figsize=[15,5])
fig.subplots_adjust(hspace=0.4)
xlabel = 'Distance from radar (km)'
ylabel = 'Distance above radar (km)'
colorbar_label = 'Equivalent reflectivity factor (dBZ)'
#nplots = radar.nsweeps

fixed_angle = radar.fixed_angle['data'][0]
title = 'HSRHI Az=%.3f' % (fixed_angle)
ax = fig.add_subplot(1, 1, 1)
display.plot('reflectivity_horizontal', 0, vmin=-20, vmax=20,
              mask_outside=True, title="RHI",
              axislabels=(xlabel, ylabel),
              cmap='pyart_HomeyerRainbow',
              colorbar_label=colorbar_label, ax=ax)
display.set_limits(ylim=[0, 15], ax=ax)

plt.show()



""" Unit Tests for Py-ART's util/radar_utils.py module. """

import numpy as np
import pyart


def test_is_vpt():
    radar = pyart.testing.make_empty_ppi_radar(10, 36, 3)
    assert not pyart.util.is_vpt(radar)
    pyart.util.to_vpt(radar)
    assert pyart.util.is_vpt(radar)

def test_join_radar():
    radar1 = pyart.testing.make_empty_ppi_radar(10, 36, 3)
    radar1.instrument_parameters = {
        'nyquist_velocity': {'data': np.array([6] * 3)}
    }
    field = {'data': np.ones((36 * 3, 10))}
    radar1.add_field('f1', field)
    radar1.add_field('f2', field)
    radar2 = pyart.testing.make_empty_ppi_radar(10, 36, 4)
    radar2.instrument_parameters = {
        'nyquist_velocity': {'data': np.array([8] * 4)}
    }
    field = {'data': np.ones((36 * 4, 10))}
    radar2.add_field('f1', field)

    radar3 = pyart.util.join_radar(radar1, radar2)
    assert 'f1' in radar3.fields
    assert len(radar3.instrument_parameters['nyquist_velocity']['data']) == 7

def test_to_vpt():
    # single scan
    radar = pyart.testing.make_empty_ppi_radar(10, 36, 3)
    radar.instrument_parameters = {
        'prt_mode': {'data': np.array(['fixed'] * 3)}
    }
    pyart.util.to_vpt(radar)
    assert pyart.util.is_vpt(radar)
    assert radar.nsweeps == 1
    assert radar.azimuth['data'][10] == 0.0
    assert radar.elevation['data'][0] == 90.0
    assert len(radar.instrument_parameters['prt_mode']['data']) == 1

    # multiple scans
    radar = pyart.testing.make_empty_ppi_radar(10, 36, 3)
    radar.instrument_parameters = {
        'prt_mode': {'data': np.array(['fixed'] * 3)}
    }
    pyart.util.to_vpt(radar, False)
    assert pyart.util.is_vpt(radar)
    assert radar.nsweeps == 108
    assert radar.azimuth['data'][10] == 10.0
    assert radar.elevation['data'][0] == 90.0
    assert len(radar.instrument_parameters['prt_mode']['data']) == 108

def test_subset_radar():
    radar = pyart.testing.make_empty_ppi_radar(10, 36, 3)
    field = {'data': np.ones((36 * 3, 10))}
    radar.add_field('f1', field)
    radar.add_field('f2', field)
    azi_min = 10
    azi_max = 100
    rng_min = 200
    rng_max = 800
    ele_min = 0.75
    ele_max = 0.75
    radarcut = pyart.util.radar_utils.subset_radar(radar, ['f1'], rng_min = 200,
                                                rng_max = 800, ele_min = 0.75,
                                                ele_max = 0.75, azi_min = 10,
                                                azi_max = 100)
    # assert correct domain and correct fields
    assert(radarcut.azimuth['data'].min() >= azi_min)
    assert(radarcut.azimuth['data'].max() <= azi_max)
    assert(radarcut.range['data'].min() >= rng_min)
    assert(radarcut.range['data'].max() <= rng_max)
    assert(radarcut.elevation['data'].min() >= ele_min)
    assert(radarcut.elevation['data'].max() <= ele_max)
    assert(list(radarcut.fields) == ['f1'])
    
# read in example file
radar = pyart.io.read_nexrad_archive(pyart.testing.NEXRAD_ARCHIVE_MSG31_FILE)
def test_image_mute_radar():

    # image mute example file
    mute_radar = pyart.util.image_mute_radar(
        radar, field='reflectivity', mute_field='cross_correlation_ratio',
        mute_threshold=0.97)

    # check that fields are added to the radar object
    assert 'nonmuted_reflectivity' in mute_radar.fields.keys()
    assert 'muted_reflectivity' in mute_radar.fields.keys()

    # check that the number of points in the muted and non muted reflectivity
    # fields have the same number of points as the rhoHV field
    n_rhohv = np.sum(~mute_radar.fields['cross_correlation_ratio']['data'].mask)
    n_mutez = np.sum(~mute_radar.fields['muted_reflectivity']['data'].mask)
    n_nonmutez = np.sum(~mute_radar.fields['nonmuted_reflectivity']['data'].mask)

    assert n_mutez + n_nonmutez == n_rhohv



""" Unit Tests for Py-ART's retrieve/comp_z.py module. """

import copy

import numpy as np
from numpy.testing import assert_equal, assert_array_equal

import pyart


def test_composite_z():
    # initalize test radar object
    radar = pyart.io.read(pyart.testing.NEXRAD_ARCHIVE_MSG31_FILE)
    ref_field = 'reflectivity'
    gatefilter = pyart.filters.GateFilter(radar)
    gatefilter.exclude_transition()
    gatefilter.exclude_below('cross_correlation_ratio', 0.95)

    ################# Trivial first test, all 0s ################

    # make fake radar data with 0 everywhere
    z = np.zeros(radar.fields['reflectivity']['data'].shape)
    z_old = radar.fields['reflectivity']

    z_new = copy.deepcopy(z_old)
    z_new['data'] = z.astype("float32")
    radar.add_field('reflectivity', z_new, replace_existing=True)

    compz = pyart.retrieve.composite_reflectivity(
        radar, field=ref_field, gatefilter=gatefilter)
    assert_equal(compz.fields['composite_reflectivity']['data'].max(), 0)
    
    ################# Second test ################
    # Insert 1 layer of all 40 dBZ
    z = np.zeros(radar.fields['reflectivity']['data'].shape)

    # choose random sweep, it shouldnt matter which one. 
    random_sweep = np.random.randint(0, radar.nsweeps)
    # loop over all measured sweeps 
    for sweep in radar.sweep_number['data']:
        if sweep == random_sweep:
            #get start and stop index numbers
            s_idx = radar.sweep_start_ray_index['data'][sweep]
            e_idx = radar.sweep_end_ray_index['data'][sweep] + 1
            z[s_idx:e_idx, :] = 40 

    z_new = copy.deepcopy(z_old)
    z_new['data'] = z.astype("float32")
    radar.add_field('reflectivity', z_new, replace_existing=True)
    compz = pyart.retrieve.composite_reflectivity(
        radar, field=ref_field, gatefilter=gatefilter)
    assert_equal(compz.fields['composite_reflectivity']['data'].max(), 40)
   
    ################# Third test ################
    # Have dBZ increase according to range bin 
    z = np.zeros(radar.fields['reflectivity']['data'].shape)

    # choose random sweep, it shouldnt matter which one. 
    random_sweep = np.random.randint(0,radar.nsweeps)

    # loop over all measured sweeps 
    for sweep in radar.sweep_number['data']:
        if sweep == random_sweep:
            # get start and stop index numbers
            s_idx = radar.sweep_start_ray_index['data'][sweep]
            e_idx = radar.sweep_end_ray_index['data'][sweep] + 1
            z[s_idx:e_idx, :] = np.tile(
                np.arange(0, z.shape[1])[np.newaxis, :], (e_idx - s_idx, 1))

    z_new = copy.deepcopy(z_old)
    z_new['data'] = z.astype("float32")
    radar.add_field('reflectivity', z_new, replace_existing=True)
    compz = pyart.retrieve.composite_reflectivity(
        radar, field=ref_field, gatefilter=gatefilter)

    # choose a random az
    random_az = np.random.randint(0, 720)
    assert_array_equal(compz.fields['composite_reflectivity']['data'][random_az, :],
                       np.arange(0, z.shape[1]))



"""
====================================
Extract a radar column above a point
====================================

Given a radar and a point, extract the column of radar data values above
a point
"""

# Author: Maxwell Grover (mgrover@anl.gov)
# License: BSD 3 clause

import pyart
import numpy as np
import matplotlib.pyplot as plt
import cartopy.crs as ccrs
from pyart.testing import get_test_data

# Read in some test data
filename = get_test_data('swx_20120520_0641.nc')
radar = pyart.io.read(filename)

######################################
# **Plot the first sweep and our desired point**
#
# Let's visualize our radar data from a single sweep, and plot
# the location of our desired point on a map.
# This will provide some context as to where we are extracting our
# column of values.

site_lon = -97.73 # longitude in degrees
site_lat = 36.41 # latitdue in degrees

# Setup the RadarMapDisplay and add our projection
display = pyart.graph.RadarMapDisplay(radar)
ax = plt.subplot(111, projection=ccrs.PlateCarree())

# Visualize the reflectivity field, using the lowest sweep with
# latitude and longitude lines
display.plot_ppi_map('reflectivity_horizontal',
                     0, ax=ax,vmin=-32, vmax=64.,
                     lon_lines=np.arange(-98, -97, .2),
                     lat_lines=np.arange(36, 37, .2))

# Plot our site location on top of the radar image
ax.scatter(site_lon,
           site_lat,
           color='black');

######################################
# Now that we have our point defined, and our radar object, we can use the following 
# utility function in Py-ART to subset a column
ds = pyart.util.columnsect.get_field_location(radar, site_lat, site_lon)

######################################
# This function returns an xarray dataset, with all of our data fields!
print(ds)

######################################
# **Visualize the Reflectivity Values in the Column**
#
# Let's visualize the reflectivity values in the column
# above our point, which is stored in our new dataset
ds.corrected_reflectivity_horizontal.plot(y='height');
